/// System includes
#include <proto/muimaster.h>
#include <libraries/mui.h>
#include <workbench/workbench.h>
#include <proto/exec.h>
#include <intuition/intuition.h>
#include <exec/libraries.h>
#include <proto/intuition.h>
#include <stdio.h>
#include <string.h>

#include <proto/alib.h>
#include <proto/utility.h>
////

#include <private/vapor/vapor.h>
#include "util.h"
#include "application.h"
#include "pagenumberclass.h"
#include "system/functions.h"

#define TYPE_END    0
#define TYPE_NUMBER 1
#define TYPE_RANGE  2
 

typedef struct
{
  char type;
  int min;
  int max;
  int page_number;
}
PageNumberItem;


struct Data
{
	Object *str_pages;
	char reverse;
	int end;
	int max;
	int min;
	int pos;
	int sub_pos;
//	int skip_mode;
	PageNumberItem pages_bitmap[1024];
};

#define D(x)

DEFNEW
{
	Object *str_pages;
	char initial_range[128];

	if ((obj = DoSuperNew(cl, obj,
		
		Child, str_pages = StringObject, 
			 MUIA_String_Accept, "0123456789-, ",
			MUIA_Frame, MUIV_Frame_String,
		End,
	TAG_MORE, INITTAGS)))
	{
		GETDATA;
		
		data->min = GetTagData(MUIA_VPDFNumberGeneratorPageStart, 1, INITTAGS);
		data->max = GetTagData(MUIA_VPDFNumberGeneratorPageEnd, 999999999, INITTAGS); 
		data->reverse = FALSE;
		data->str_pages = str_pages;
		
		memset(data->pages_bitmap, 0, sizeof(data->pages_bitmap));
		data->pos = 0;
		data->sub_pos = -1;  // position within range
	}
	
	return (ULONG)obj;
}


DEFDISP
{
	GETDATA;
		
	return DOSUPER;
}

DEFMMETHOD(VPDFNumberGeneratorFirst)
{
	GETDATA;
	char *str;
	int total_pages = 0;	
	char *break_pos, *pch;
	int current_pos = 0;
	char buffer[1024];
	data->pos = 0;
	data->sub_pos = -1;  // position within range
		
	get(data->str_pages, MUIA_String_Contents, &str);
	strcpy(buffer, str);
	
	pch = strtok (buffer, ",");
	
	while ((pch != NULL))
	{
		// checking if we have range here
		if ((break_pos = (strchr(pch, '-'))))
		{

			data->pages_bitmap[current_pos].type = TYPE_RANGE; 
			data->pages_bitmap[current_pos].max  = atoi(break_pos+1); 
			data->pages_bitmap[current_pos].min  = atoi(pch);
			
			//reverse range if min>max
			if (data->pages_bitmap[current_pos].max<data->pages_bitmap[current_pos].min)
			{
				int tmp;
				tmp = data->pages_bitmap[current_pos].min;
				data->pages_bitmap[current_pos].min = data->pages_bitmap[current_pos].max;
				data->pages_bitmap[current_pos].max = tmp;
				
			}
			
			// one page range
			if (data->pages_bitmap[current_pos].min==data->pages_bitmap[current_pos].max)
			{
				data->pages_bitmap[current_pos].type = TYPE_NUMBER; 
				data->pages_bitmap[current_pos].page_number = data->pages_bitmap[current_pos].max;
				
				if (data->pages_bitmap[current_pos].page_number>data->max)
				{
					data->pages_bitmap[current_pos].page_number = data->max;
				}
				total_pages++;
			}
			else 
			{
				int pages;
				if (data->pages_bitmap[current_pos].max>data->max)
				{
					data->pages_bitmap[current_pos].max = data->max;
				}
				pages = data->pages_bitmap[current_pos].max-data->pages_bitmap[current_pos].min+1;
				total_pages += pages;
				
				if (msg->booklet)
				{
					int i;
					for (i=0; i<pages; i++)
					{
						data->pages_bitmap[current_pos+i].type = TYPE_NUMBER; 
						data->pages_bitmap[current_pos+i].page_number = data->pages_bitmap[current_pos].min+i;
					}
					current_pos += total_pages-1;
				}
			}
		
			D(KPrintF("range [%d, %d]\n", data->pages_bitmap[current_pos].min, data->pages_bitmap[current_pos].max)); 
		}
		else // one page
		{
			data->pages_bitmap[current_pos].page_number = atoi(pch);
			data->pages_bitmap[current_pos].type = TYPE_NUMBER; 
			
			if (data->pages_bitmap[current_pos].page_number>data->max)
			{
				data->pages_bitmap[current_pos].page_number = data->max;
			}
			total_pages++;
			D(KPrintF("number [%d]\n", data->pages_bitmap[current_pos].page_number)); 
		}
		current_pos++;
		
		pch = strtok (NULL, ",");
	}
	
	data->pages_bitmap[current_pos].type = TYPE_END;
	
	// if we have reverse mode turned on: start from end
	if (msg->reverse == TRUE)
	{
		data->pos = current_pos-1;
		data->reverse = TRUE;
	}
	else
	{
		data->reverse = FALSE;
	}
		
	D(KPrintF("before booklet total pages: %d\n", total_pages));
	
	
	if (msg->booklet)
	{
		int i, j, old_total_pages;
		old_total_pages = total_pages;
		total_pages = ((total_pages+3)/4)*4;
		
		if (old_total_pages<total_pages)
		{
			for (i=old_total_pages; i<total_pages; i++, current_pos++)
			{
				data->pages_bitmap[current_pos].type = TYPE_NUMBER; 
				data->pages_bitmap[current_pos].page_number = 0;
			}
			data->pages_bitmap[current_pos].type = TYPE_END;
		}
		
		PageNumberItem pages_bitmap_temp[1024];
		
//		1   2   3   4   5   6   7   8  
//      8   1   2   7   6   3   4   5  
//     last 0 1 last-1 last-2 2 3 last-3
		D(KPrintF("booklet total pages: %d\n", total_pages));
		for (i=0,j=0; i<total_pages; i+=4, j+=2)
		{
			pages_bitmap_temp[i]   = data->pages_bitmap[total_pages-1-j];
			D(KPrintF("booklet 1 [%d]: %d\n", i, pages_bitmap_temp[i].page_number ));
			
			pages_bitmap_temp[i+1] = data->pages_bitmap[j];
			D(KPrintF("booklet 2 [%d]: %d\n", i+1, pages_bitmap_temp[i+1].page_number ));
			
		    pages_bitmap_temp[i+2] = data->pages_bitmap[j+1];
			D(KPrintF("booklet 3 [%d]: %d\n", i+2, pages_bitmap_temp[i+2].page_number ));
			
			pages_bitmap_temp[i+3] = data->pages_bitmap[total_pages-2-j];
			D(KPrintF("booklet 4 [%d]: %d\n", i+3, pages_bitmap_temp[i+3].page_number ));
		}	
		
		CopyMem(pages_bitmap_temp, data->pages_bitmap, sizeof(PageNumberItem)*1024);
	}


	return total_pages; 
}


DEFMMETHOD(VPDFNumberGeneratorNext)
{
	GETDATA;

	if(!data->reverse)
	{
		switch(data->pages_bitmap[data->pos].type)
		{
		case TYPE_NUMBER:
			data->pos++;
			return data->pages_bitmap[data->pos - 1].page_number;
			break;
		case TYPE_RANGE:
			if(data->sub_pos == -1) // start of range
			{
				data->sub_pos = data->pages_bitmap[data->pos].min;
				return data->sub_pos;
			}
			else
			{
				data->sub_pos++;

				if(data->sub_pos == data->pages_bitmap[data->pos].max)
				{
					data->pos++;
					data->sub_pos = -1;
					return data->pages_bitmap[data->pos - 1].max;
				}
				return data->sub_pos;
			}
		break;
		default:
			return MUIV_VPDFNumberGeneratorEndPos;  // end of page list
		break;
		}
	}
	else // reverse mode
	{
		if (data->pos>=0)
		{
			switch(data->pages_bitmap[data->pos].type)
			{
				case TYPE_NUMBER:
					data->pos--;
					return data->pages_bitmap[data->pos + 1].page_number;
				break;
				case TYPE_RANGE:
					if(data->sub_pos == -1)  // start of range
					{
						data->sub_pos = data->pages_bitmap[data->pos].max;
						return data->sub_pos;
					}
					else
					{
						data->sub_pos--;

						if(data->sub_pos == data->pages_bitmap[data->pos].min)
						{
							data->pos--;
							data->sub_pos = -1;
							return data->pages_bitmap[data->pos + 1].min;
						}
						return data->sub_pos;
					}
				break;
			}
		}
		else
			return MUIV_VPDFNumberGeneratorEndPos;
	}
	return MUIV_VPDFNumberGeneratorEndPos;
}


DEFSET
{
	char initial_range[128];
	GETDATA;

	FORTAG(INITTAGS)
	{
		/*case MUIA_VPDFNumberGeneratorPageSkip:
		{
			data->skip_mode = tag->ti_Data;
			break;
		}
		*/
		
		case MUIA_VPDFNumberGeneratorPageStart: 
			data->min = tag->ti_Data;
			sprintf(initial_range, "%d - %d", data->min, data->end);
			set(data->str_pages, MUIA_String_Contents, initial_range);
		break; 
		
		case MUIA_VPDFNumberGeneratorPageEnd:
			data->end = tag->ti_Data;
			sprintf(initial_range, "%d - %d", data->min, data->end);
			set(data->str_pages, MUIA_String_Contents, initial_range);
		break;
		
		case MUIA_VPDFNumberGeneratorPageMax:
			data->max = tag->ti_Data;
		break;
	}
	
	NEXTTAG

	return DOSUPER;
}



BEGINMTABLE
	DECNEW
	DECDISP
	DECSET
	DECMMETHOD(VPDFNumberGeneratorFirst)
	DECMMETHOD(VPDFNumberGeneratorNext)
ENDMTABLE

DECSUBCLASS_NC(MUIC_Group, VPDFNumberGeneratorClass)